# -*- python -*-
# $RCSfile: workerGUI.py,v $
# $Revision: 1.21.8.1 $
# $Author: langer $
# $Date: 2008-07-03 21:34:55 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from ooflib.common import debug
from ooflib.common import parallel_enable
from ooflib.common import mainthread
from ooflib.common import thread_enable
from ooflib.common import worker
from ooflib.common.IO import threadmanager

# GUIThreadedWorkerBlock is called when running a script in GUI mode.
# The script is running on a subthread already, and commands in the
# script need to be executed sequentially, so we have to wait for each
# one to finish.

class GUIThreadedWorkerBlock(worker.ThreadedWorker):
    def meanwhile(self):
        self.worker.join()              # waits for child thread to finish


## parallel worker classes start here

class GUIParallelThreadedWorker(worker.ThreadedWorker,
                                worker.ParallelWork):
    def __init__ (self, menuitem, args, kwargs):
        self.worker = worker.ParallelThreadedWorkerCore(menuitem, args, kwargs)
        threadmanager.threadManager.newWorker(self.worker)
        worker.ParallelWork.__init__(self, menuitem, args, kwargs)
        ## TODO: The class hierarchy is messed up... This ought to
        ## call ThreadedWorker.__init__, but it can't because
        ## ThreadedWorker.__init__ also sets self.worker.  self.worker
        ## should be set by a virtual function *other* than __init__.
        


class GUIParallelThreadedWorkerBlock(GUIThreadedWorkerBlock,
                                     worker.ParallelWork):
    def __init__ (self, menuitem, args, kwargs):
        self.worker = worker.ParallelThreadedWorkerCore(
            menuitem, args, kwargs)
        threadmanager.threadManager.newWorker(self.worker)
        worker.ParallelWork.__init__(self, menuitem, args, kwargs)
        
## parallel worker classes end here


def getThreadedWorker(menuitem, argtuple, argdict):
    if not thread_enable.query():
        ## serial workers
        return worker.NonThreadedWorker(menuitem, argtuple, argdict)
    if mainthread.mainthread():
        return worker.ThreadedWorker(menuitem, argtuple, argdict)
    return GUIThreadedWorkerBlock(menuitem, argtuple, argdict)

worker.getThreadedWorker = getThreadedWorker

def getThreadedParallelWorker(menuitem, argtuple, argdict):
    ## parallel workers
    ## only called by front-end
    ## back-end is running on text-mode, without any text
    ## progress bars
    if not thread_enable.query():
        ## if GUI is not in place yet, do not 
        ## launch a thread. Instead run a serial
        ## parallel worker
        return worker.ParallelWorker(menuitem, argtuple, argdict)
    if mainthread.mainthread():
        ## if main thread launches a parallel thread
        ## launch a GUI worker
        return GUIParallelThreadedWorker(menuitem, argtuple, argdict)
    ## if subthread is attempting to launch a thread,
    ## block the subthread, until, sub-subthread joins
    return GUIParallelThreadedWorkerBlock(menuitem, argtuple, argdict)


worker.getThreadedParallelWorker = getThreadedParallelWorker
