// -*- C++ -*-
// $RCSfile: elasticity.h,v $
// $Revision: 1.22 $
// $Author: vrc $
// $Date: 2007-09-26 22:03:17 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifndef ELASTICITY_H
#define ELASTICITY_H

#include <Python.h>
#include "engine/property.h"
#include "cijkl.h"
#include <string>

class CSubProblem;
class Element;
class ElementNodeIterator;
class FEMesh;
class Flux;
class FluxData;
class Material;
class OutputVal;
class Position;
class PropertyOutput;
class SymmetricTensorFlux;
class TwoVectorField;

class Elasticity : public Property {
public:
  Elasticity(const std::string &name, PyObject *registry);
  virtual ~Elasticity() {}
  virtual void precompute(FEMesh*);
  virtual void fluxmatrix(const FEMesh*, const Element*, 
			  const ElementFuncNodeIterator &,
			  const Flux*, FluxData*,
			  const MasterPosition&) const;
  virtual int integration_order(const CSubProblem*, const Element*) const;
  virtual bool constant_in_space() const { return true; }

  // elastic modulus to be defined in subclasses
  virtual const Cijkl cijkl(const FEMesh*, const Element*,
			    const MasterPosition&) const = 0;

  virtual void output(const FEMesh*, const Element*, const PropertyOutput*,
		      const MasterPosition&, OutputVal*) const;
  virtual bool is_symmetric(const CSubProblem*) const { return true; }
protected:
  TwoVectorField *displacement;
  SymmetricTensorFlux *stress_flux;
};

#endif
